/// changed this line: lowestAddress = 0x7E00;//I HAD TO FORCE THIS FOR THE ATMEGA328P

void showHex (const byte b, const boolean newline, const boolean show0x)
{
  if (show0x)
    Serial.print (F("0x"));
  // try to avoid using sprintf
  char buf [4];
  buf [0] = ((b >> 4) & 0x0F) | '0';
  buf [1] = (b & 0x0F) | '0';
  buf [2] = ' ';
  buf [3] = 0;
  if (buf [0] > '9')
    buf [0] += 7;
  if (buf [1] > '9')
    buf [1] += 7;
  Serial.print (buf);
  if (newline)
    Serial.println ();
}  // end of showHex
// blink one or two LEDs for "times" times, with a delay of "interval". Wait a second and do it again "repeat" times.
void blink (const int whichLED1,
            const int whichLED2,
            const byte times = 1,
            const unsigned long repeat = 1,
            const unsigned long interval = 200)
{
  for (unsigned long i = 0; i < repeat; i++)
  {
    for (byte j = 0; j < times; j++)
    {
      digitalWrite (whichLED1, HIGH);
      if (whichLED2 != noLED)
        digitalWrite (whichLED2, HIGH);
      delay (interval);
      digitalWrite (whichLED1, LOW);
      if (whichLED2 != noLED)
        digitalWrite (whichLED2, LOW);
      delay (interval);
    }  // end of for loop
    if (i < (repeat - 1))
      delay (1000);
  }  // end of repeating the sequence
}  // end of blink

void ShowMessage (const byte which)
{
  // first turn off all LEDs
  digitalWrite (errorLED, LOW);
  digitalWrite (workingLED, LOW);
  digitalWrite (readyLED, LOW);
  Serial.println("SOMETHING IS NOT RIGHT!! Press RED Button to try again...");
  while (1) {
    // now flash an appropriate sequence
    switch (which)
    {
      // problems with SD card or finding the file
      case MSG_NO_SD_CARD:                      blink (errorLED, noLED, 1, 5); break;
      case MSG_CANNOT_OPEN_FILE:                blink (errorLED, noLED, 2, 5); break;

      // problems reading the .hex file
      case MSG_LINE_TOO_LONG:                   blink (errorLED, workingLED, 1, 5); break;
      case MSG_LINE_TOO_SHORT:                  blink (errorLED, workingLED, 2, 5); break;
      case MSG_LINE_DOES_NOT_START_WITH_COLON:  blink (errorLED, workingLED, 3, 5); break;
      case MSG_INVALID_HEX_DIGITS:              blink (errorLED, workingLED, 4, 5); break;
      case MSG_BAD_SUMCHECK:                    blink (errorLED, workingLED, 5, 5); break;
      case MSG_LINE_NOT_EXPECTED_LENGTH:        blink (errorLED, workingLED, 6, 5); break;
      case MSG_UNKNOWN_RECORD_TYPE:             blink (errorLED, workingLED, 7, 5); break;
      case MSG_NO_END_OF_FILE_RECORD:           blink (errorLED, workingLED, 8, 5); break;

      // problems with the file contents
      case MSG_FILE_TOO_LARGE_FOR_FLASH:        blink (errorLED, workingLED, 9, 5); break;

      // problems programming the chip
      case MSG_CANNOT_ENTER_PROGRAMMING_MODE:  blink (errorLED, noLED, 3, 5); break;
      case MSG_NO_BOOTLOADER_FUSE:             blink (errorLED, noLED, 4, 5); break;
      case MSG_CANNOT_FIND_SIGNATURE:          blink (errorLED, noLED, 5, 5); break;
      case MSG_UNRECOGNIZED_SIGNATURE:         blink (errorLED, noLED, 6, 5); break;
      case MSG_BAD_START_ADDRESS:              blink (errorLED, noLED, 7, 5); break;
      case MSG_VERIFICATION_ERROR:             blink (errorLED, noLED, 8, 5); break;
      case MSG_FLASHED_OK:                     blink (readyLED, noLED, 3, 2); break;

      default:                                  blink (errorLED, 10, 10);  break;   // unknown error
    }  // end of switch on which message
  }
}  // end of ShowMessage

// Bit Banged SPI transfer
byte BB_SPITransfer (byte c)
{
  byte bit;

  for (bit = 0; bit < 8; bit++)
  {
    // write MOSI on falling edge of previous clock
    if (c & 0x80)
      BB_MOSI_PORT |= bit (BB_MOSI_BIT);
    else
      BB_MOSI_PORT &= ~bit (BB_MOSI_BIT);
    c <<= 1;

    // read MISO
    c |= (BB_MISO_PORT & bit (BB_MISO_BIT)) != 0;

    // clock high
    BB_SCK_PORT |= bit (BB_SCK_BIT);

    // delay between rise and fall of clock
    delayMicroseconds (BB_DELAY_MICROSECONDS);

    // clock low
    BB_SCK_PORT &= ~bit (BB_SCK_BIT);

    // delay between rise and fall of clock
    delayMicroseconds (BB_DELAY_MICROSECONDS);
  }

  return c;
}  // end of BB_SPITransfer

// execute one programming instruction ... b1 is command, b2, b3, b4 are arguments
//  processor may return a result on the 4th transfer, this is returned.
byte program (const byte b1, const byte b2 = 0, const byte b3 = 0, const byte b4 = 0)
{
  noInterrupts ();
  BB_SPITransfer (b1);
  BB_SPITransfer (b2);
  BB_SPITransfer (b3);
  byte b = BB_SPITransfer (b4);
  interrupts ();
  return b;
} // end of program

// read a byte from flash memory
byte readFlash (unsigned long addr)
{
  byte high = (addr & 1) ? 0x08 : 0;  // set if high byte wanted
  addr >>= 1;  // turn into word address

  // set the extended (most significant) address byte if necessary
  byte MSB = (addr >> 16) & 0xFF;
  if (MSB != lastAddressMSB)
  {
    program (loadExtendedAddressByte, 0, MSB);
    lastAddressMSB = MSB;
  }  // end if different MSB

  return program (readProgramMemory | high, highByte (addr), lowByte (addr));
} // end of readFlash
// write a byte to the flash memory buffer (ready for committing)

void writeFlash (unsigned long addr, const byte data)
{
  byte high = (addr & 1) ? 0x08 : 0;  // set if high byte wanted
  addr >>= 1;  // turn into word address
  program (loadProgramMemory | high, 0, lowByte (addr), data);
} // end of writeFlash

// convert two hex characters into a byte
//    returns true if error, false if OK
bool hexConv (const char * (& pStr), byte & b)
{

  if (!isxdigit (pStr [0]) || !isxdigit (pStr [1]))
  {
    ShowMessage (MSG_INVALID_HEX_DIGITS);
    return true;
  } // end not hex

  b = *pStr++ - '0';
  if (b > 9)
    b -= 7;

  // high-order nybble
  b <<= 4;

  byte b1 = *pStr++ - '0';
  if (b1 > 9)
    b1 -= 7;

  b |= b1;

  return false;  // OK
}  // end of hexConv
// poll the target device until it is ready to be programmed
void pollUntilReady ()
{
  if (currentSignature.timedWrites)
    delay (10);  // at least 2 x WD_FLASH which is 4.5 mS
  else
  {
    while ((program (pollReady) & 1) == 1)
    {}  // wait till ready
  }  // end of if
}  // end of pollUntilReady

// shows progress, toggles working LED
void showProgress ()
{
  digitalWrite (workingLED, ! digitalRead (workingLED));
}  // end of showProgress
// clear entire temporary page to 0xFF in case we don't write to all of it
void clearPage ()
{
  unsigned int len = currentSignature.pageSize;
  for (unsigned int i = 0; i < len; i++)
    writeFlash (i, 0xFF);
}  // end of clearPage
// commit page to flash memory
void commitPage (unsigned long addr)
{
  addr >>= 1;  // turn into word address

  // set the extended (most significant) address byte if necessary
  byte MSB = (addr >> 16) & 0xFF;
  if (MSB != lastAddressMSB)
  {
    program (loadExtendedAddressByte, 0, MSB);
    lastAddressMSB = MSB;
  }  // end if different MSB

  showProgress ();

  program (writeProgramMemory, highByte (addr), lowByte (addr));
  pollUntilReady ();

  clearPage();  // clear ready for next page full
}  // end of commitPage
// write data to temporary buffer, ready for committing
void writeData (const unsigned long addr, const byte * pData, const int length)
{
  // write each byte
  for (int i = 0; i < length; i++)
  {
    unsigned long thisPage = (addr + i) & pagemask;
    // page changed? commit old one
    if (thisPage != oldPage && oldPage != NO_PAGE)
      commitPage (oldPage);
    // now this is the current page
    oldPage = thisPage;
    // put byte into work buffer
    writeFlash (addr + i, pData [i]);
  }  // end of for

}  // end of writeData
void verifyData (const unsigned long addr, const byte * pData, const int length)
{
  // check each byte
  for (int i = 0; i < length; i++)
  {
    unsigned long thisPage = (addr + i) & pagemask;
    // page changed? show progress
    if (thisPage != oldPage && oldPage != NO_PAGE)
      showProgress ();
    // now this is the current page
    oldPage = thisPage;

    byte found = readFlash (addr + i);
    byte expected = pData [i];
    if (found != expected)
      errors++;
  }  // end of for

}  // end of verifyData

/*
  Line format:

  :nnaaaatt(data)ss

  Where:
  :      = a colon

  (All of below in hex format)

  nn     = length of data part
  aaaa   = address (eg. where to write data)
  tt     = transaction type
           00 = data
           01 = end of file
           02 = extended segment address (changes high-order byte of the address)
           03 = start segment address *
           04 = linear address *
           05 = start linear address *
  (data) = variable length data
  ss     = sumcheck

              We don't use these

*/

// returns true if error, false if OK
bool processLine (const char * pLine, const byte action)
{
  if (*pLine++ != ':')
  {
    ShowMessage (MSG_LINE_DOES_NOT_START_WITH_COLON);
    return true;  // error
  }

  const int maxHexData = 40;
  byte hexBuffer [maxHexData];
  int bytesInLine = 0;

  if (action == checkFile)
    if (lineCount++ % 40 == 0)
      showProgress ();

  // convert entire line from ASCII into binary
  while (isxdigit (*pLine))
  {
    // can't fit?
    if (bytesInLine >= maxHexData)
    {
      ShowMessage (MSG_LINE_TOO_LONG);
      return true;
    } // end if too long

    if (hexConv (pLine, hexBuffer [bytesInLine++]))
      return true;
  }  // end of while

  if (bytesInLine < 5)
  {
    ShowMessage (MSG_LINE_TOO_SHORT);
    return true;
  }

  // sumcheck it

  byte sumCheck = 0;
  for (int i = 0; i < (bytesInLine - 1); i++)
    sumCheck += hexBuffer [i];

  // 2's complement
  sumCheck = ~sumCheck + 1;

  // check sumcheck
  if (sumCheck != hexBuffer [bytesInLine - 1])
  {
    ShowMessage (MSG_BAD_SUMCHECK);
    return true;
  }

  // length of data (eg. how much to write to memory)
  byte len = hexBuffer [0];

  // the data length should be the number of bytes, less
  //   length / address (2) / transaction type / sumcheck
  if (len != (bytesInLine - 5))
  {
    ShowMessage (MSG_LINE_NOT_EXPECTED_LENGTH);
    return true;
  }

  // two bytes of address
  unsigned long addrH = hexBuffer [1];
  unsigned long addrL = hexBuffer [2];

  unsigned long addr = addrL | (addrH << 8);

  byte recType = hexBuffer [3];

  switch (recType)
  {
    // stuff to be written to memory
    case hexDataRecord:
      //lowestAddress  = min (lowestAddress, addr + extendedAddress);
      lowestAddress = 0x7E00;//I HAD TO FORCE THIS FOR THE ATMEGA328P
      highestAddress = max (lowestAddress, addr + extendedAddress + len - 1);
      bytesWritten += len;

      switch (action)
      {
        case checkFile:  // nothing much to do, we do the checks anyway
          break;

        case verifyFlash:
          verifyData (addr + extendedAddress, &hexBuffer [4], len);
          break;

        case writeToFlash:
          writeData (addr + extendedAddress, &hexBuffer [4], len);
          break;
      } // end of switch on action
      break;

    // end of data
    case hexEndOfFile:
      gotEndOfFile = true;
      break;

    // we are setting the high-order byte of the address
    case hexExtendedSegmentAddressRecord:
      extendedAddress = ((unsigned long) hexBuffer [4]) << 12;
      break;

    // ignore these, who cares?
    case hexStartSegmentAddressRecord:
    case hexExtendedLinearAddressRecord:
    case hexStartLinearAddressRecord:
      break;

    default:
      ShowMessage (MSG_UNKNOWN_RECORD_TYPE);
      return true;
  }  // end of switch on recType

  return false;
} // end of processLine

//------------------------------------------------------------------------------
// returns true if error, false if OK
bool readHexFile (const char * fName, const byte action)
{
  const int maxLine = 80;
  char buffer[maxLine];
  ifstream sdin (fName);
  int lineNumber = 0;
  gotEndOfFile = false;
  extendedAddress = 0;
  errors = 0;
  lowestAddress = 0xFFFFFFFF;
  //lowestAddress
  highestAddress = 0;
  bytesWritten = 0;
  progressBarCount = 0;

  pagesize = currentSignature.pageSize;
  pagemask = ~(pagesize - 1);
  oldPage = NO_PAGE;

  // check for open error
  if (!sdin.is_open())
  {
    ShowMessage (MSG_CANNOT_OPEN_FILE);
    return true;
  }

  switch (action)
  {
    case checkFile:
      break;

    case verifyFlash:
      break;

    case writeToFlash:
      program (progamEnable, chipErase);   // erase it
      delay (20);  // for Atmega8
      pollUntilReady ();
      clearPage();  // clear temporary page
      break;
  } // end of switch

  while (sdin.getline (buffer, maxLine))
  {
    lineNumber++;
    int count = sdin.gcount();
    if (sdin.fail())
    {
      ShowMessage (MSG_LINE_TOO_LONG);
      return true;
    }  // end of fail (line too long?)

    // ignore empty lines
    if (count > 1)
    {
      if (processLine (buffer, action))
      {
        return true;  // error
      }
    }
  }    // end of while each line

  if (!gotEndOfFile)
  {
    ShowMessage (MSG_NO_END_OF_FILE_RECORD);
    return true;
  }

  switch (action)
  {
    case writeToFlash:
      // commit final page
      if (oldPage != NO_PAGE)
        commitPage (oldPage);
      break;

    case verifyFlash:
      if (errors > 0)
      {
        ShowMessage (MSG_VERIFICATION_ERROR);
        return true;
      }  // end if
      break;

    case checkFile:
      break;
  }  // end of switch

  return false;
}  // end of readHexFile

// returns true if managed to enter programming mode
bool startProgramming ()
{

  byte confirm;
  pinMode (RESET, OUTPUT);
  digitalWrite (MSPIM_SCK, LOW);
  pinMode (MSPIM_SCK, OUTPUT);
  pinMode (BB_MOSI, OUTPUT);
  //pinMode (BB_MISO, INPUT);
  unsigned int timeout = 0;

  // we are in sync if we get back programAcknowledge on the third byte
  do
  {
    // regrouping pause
    delay (100);

    // ensure SCK low
    noInterrupts ();
    digitalWrite (MSPIM_SCK, LOW);
    // then pulse reset, see page 309 of datasheet
    digitalWrite (RESET, HIGH);
    delayMicroseconds (10);  // pulse for at least 2 clock cycles
    digitalWrite (RESET, LOW);
    interrupts ();

    delay (25);  // wait at least 20 mS
    //delay (25);
    noInterrupts ();
    BB_SPITransfer (progamEnable);
    BB_SPITransfer (programAcknowledge);
    confirm = BB_SPITransfer (0);
    BB_SPITransfer (0);
    interrupts ();

    if (confirm != programAcknowledge)
    {
      if (timeout++ >= ENTER_PROGRAMMING_ATTEMPTS)
        return false;
    }  // end of not entered programming mode

  } while (confirm != programAcknowledge);
  return true;  // entered programming mode OK
}  // end of startProgramming

void stopProgramming ()
{
  // turn off pull-ups
  digitalWrite (RESET, LOW);
  digitalWrite (MSPIM_SCK, LOW);
  digitalWrite (BB_MOSI, LOW);
  digitalWrite (BB_MISO, LOW);

  // set everything back to inputs
  pinMode (RESET, INPUT);
  pinMode (MSPIM_SCK, INPUT);
  pinMode (BB_MOSI, INPUT);
  pinMode (BB_MISO, INPUT);

} // end of stopProgramming

void getSignature ()
{
  foundSig = -1;
  lastAddressMSB = 0;

  byte sig [3];
  // Serial.print (F("Signature = "));




  for (byte i = 0; i < 3; i++)
  {
    sig [i] = program (readSignatureByte, 0, i);
    //   showHex (sig [i]);
  }  // end for each signature byte
  Serial.println ("");
  for (unsigned int j = 0; j < NUMITEMS (signatures); j++)
  {
    memcpy_P (&currentSignature, &signatures [j], sizeof currentSignature);

    if (memcmp (sig, currentSignature.sig, sizeof sig) == 0)
    {
      foundSig = j;
      // make sure extended address is zero to match lastAddressMSB variable
      program (loadExtendedAddressByte, 0, 0);
      return;
    }  // end of signature found
  }  // end of for each signature

  ShowMessage (MSG_UNRECOGNIZED_SIGNATURE);
}  // end of getSignature

void getFuseBytes ()
{
  fuses [lowFuse]   = program (readLowFuseByte, readLowFuseByteArg2);
  fuses [highFuse]  = program (readHighFuseByte, readHighFuseByteArg2);
  fuses [extFuse]   = program (readExtendedFuseByte, readExtendedFuseByteArg2);
  fuses [lockByte]  = program (readLockByte, readLockByteArg2);
  fuses [calibrationByte]  = program (readCalibrationByte);

  Serial.print (F("LFuse = "));
  showHex (fuses [lowFuse], true);
  Serial.print (F("HFuse = "));
  showHex (fuses [highFuse], true);
  Serial.print (F("EFuse = "));
  showHex (fuses [extFuse], true);
  Serial.print (F("Lock byte = "));
  showHex (fuses [lockByte], true);
  Serial.print (F("Clock calibration = "));
  showHex (fuses [calibrationByte], true);

}  // end of getFuseBytes


// write specified value to specified fuse/lock byte
void writeFuse (const byte newValue, const byte instruction)
{
  //  if (newValue == 0)
  //    return;  // ignore

  program (progamEnable, instruction, 0, newValue);
  pollUntilReady ();
}  // end of writeFuse

// returns true if error, false if OK
bool updateFuses (const bool writeIt)
{
  unsigned long addr;
  unsigned int  len;

  byte fusenumber = currentSignature.fuseWithBootloaderSize;

  // if no fuse, can't change it
  if (fusenumber == NO_FUSE)
  {
    //    ShowMessage (MSG_NO_BOOTLOADER_FUSE);   // maybe this doesn't matter?
    return false;  // ok return
  }

  addr = currentSignature.flashSize;
  len = currentSignature.baseBootSize;

  if (lowestAddress == 0)
  {
    // don't use bootloader
    fuses [fusenumber] |= 1;
  }
  else
  {
    byte newval = 0xFF;

    if (lowestAddress == (addr - len))
      newval = 3;
    else if (lowestAddress == (addr - len * 2))
      newval = 2;
    else if (lowestAddress == (addr - len * 4))
      newval = 1;
    else if (lowestAddress == (addr - len * 8))
      newval = 0;
    else
    {
      ShowMessage (MSG_BAD_START_ADDRESS);
      return true;
    }

    if (newval != 0xFF)
    {
      newval <<= 1;
      fuses [fusenumber] &= ~0x07;   // also program (clear) "boot into bootloader" bit
      fuses [fusenumber] |= newval;
    }  // if valid

  }  // if not address 0

  if (writeIt)
  {
    writeFuse (fuses [fusenumber], fuseCommands [fusenumber]);

  }

  return false;
}  // end of updateFuses

// returns true if error, false if OK
bool chooseInputFile ()
{

  if (readHexFile(name, checkFile))
  {
    return true;  // error, don't attempt to write
  }

  // check file would fit into device memory
  if (highestAddress > currentSignature.flashSize)
  {
    ShowMessage (MSG_FILE_TOO_LARGE_FOR_FLASH);
    return true;
  }

  // check start address makes sense
  if (updateFuses (false))
  {
    return true;
  }

  return false;
}  // end of chooseInputFile

// returns true if OK, false on error
bool writeFlashContents ()
{

  errors = 0;

  if (chooseInputFile ())
    return false;


  // ensure back in programming mode
  if (!startProgramming ())
    return false;

  // now commit to flash
  if (readHexFile(name, writeToFlash))
    return false;

  // turn ready LED on during verification
  digitalWrite (readyLED, HIGH);

  // verify
  if (readHexFile(name, verifyFlash))
    return false;

  // now fix up fuses so we can boot
  if (errors == 0)
    updateFuses (true);

  return errors == 0;
}  // end of writeFlashContents
